import json
import re
import uuid

from copy import deepcopy
from datetime import datetime, timedelta
from jsonpath_ng import parse
from .aws import s3
from .error import write_error


def load_config(event):
    """
    * Given a Cumulus message and context, returns the config object for the task
    * @param {*} event An event in the Cumulus message format with remote parts resolved
    * @param {*} context The context object passed to AWS Lambda or containing an activityArn
    * @returns {*} The task's configuration
    """
    write_error('Starting load_config')

    if 'task_config' in event:
        return event['task_config']
    return {}

def load_remote_event(event):
    """
    * Given a Cumulus message, checks for a 'replace' key and fetches a remote stored
    * object from S3 and inserts it into the configured path
    * @param {*} event An event in the Cumulus message format
    * @returns {*} A Cumulus message with the remote message resolved
    """
    write_error('Starting load_remote_event')
    if 'replace' in event:
        local_exception = event.get('exception', None)
        _s3 = s3()
        data = _s3.Object(event['replace']['Bucket'],
                          event['replace']['Key']).get()
        target_json_path = event['replace']['TargetPath']
        parsed_json_path = parse(target_json_path)
        if data is not None:
            remote_event = json.loads(data['Body'].read().decode('utf-8'))
            replacement_targets = parsed_json_path.find(event)
            if not replacement_targets or len(replacement_targets) != 1:
                raise Exception(f'Remote event configuration target {target_json_path} invalid')
            try:
                replacement_targets[0].value.update(remote_event)
            except AttributeError:
                parsed_json_path.update(event, remote_event)

            event.pop('replace')
            exception_bool = (local_exception and local_exception != 'None')
            if exception_bool and (not event['exception'] or event['exception'] == 'None'):
                event['exception'] = local_exception
    write_error('Ending load_remote_event')
    return event


# Config templating
def resolve_path_str(event, json_path_string):
    """
    * Given a Cumulus message (AWS Lambda event) and a string containing a JSONPath
    * template to interpret, returns the result of interpreting that template.
    *
    * Templating comes in three flavors:
    *   1. Single curly-braces within a string ("some{$.path}value"). The JSONPaths
    *      are replaced by the first value they match, coerced to string
    *   2. A string surrounded by double curly-braces ("{{$.path}}").  The function
    *      returns the first object matched by the JSONPath
    *   3. A string surrounded by curly and square braces ("{[$.path]}"). The function
    *      returns an array of all object matching the JSONPath
    *
    * It's likely we'll need some sort of bracket-escaping at some point down the line
    *
    * @param {*} event The Cumulus message
    * @param {*} json_path_string A string containing a JSONPath template to resolve
    * @returns {*} The resolved object
    """
    write_error('Starting resolve_path_str')

    value_regex = r"^{[^\[\]].*}$"
    array_regex = r"^{\[.*\]}$"
    template_regex = '{[^}]+}'

    if re.search(value_regex, json_path_string):
        match_data = parse(json_path_string.lstrip('{').rstrip('}')).find(event)
        return match_data[0].value if match_data else None

    if re.search(array_regex, json_path_string):
        parsed_json_path = json_path_string.lstrip('{').rstrip('}').lstrip('[').rstrip(']')
        match_data = parse(parsed_json_path).find(event)
        return [item.value for item in match_data] if match_data else []

    if re.search(template_regex, json_path_string):
        matches = re.findall(template_regex, json_path_string)
        for match in matches:
            match_data = parse(match.lstrip('{').rstrip('}')).find(event)
            if match_data:
                json_path_string = json_path_string.replace(match, match_data[0].value)
        return json_path_string

    write_error('End resolve_path_str')
    return json_path_string


def resolve_input(event, config):
    """
    * Given a Cumulus message and its config, returns the input object to send to the
    * task, as defined under config.cumulus_message
    * @param {*} event The Cumulus message
    * @param {*} config The config object
    * @returns {*} The object to place on the input key of the task's event
    """
    write_error('Starting resolve_input')
    if ('cumulus_message' in config and 'input' in config['cumulus_message']):
        input_path = config['cumulus_message']['input']
        return resolve_path_str(event, input_path)
    write_error('End resolve_input')
    return event.get('payload')

def resolve_config_templates(event, config):
    """
    * Given a config object containing possible JSONPath-templated values, resolves
    * all the values in the object using JSONPaths into the provided event.
    *
    * @param {*} event The event that paths resolve against
    * @param {*} config A config object, containing paths
    * @returns {*} A config object with all JSONPaths resolved
    """
    write_error('Starting resolve_config_templates')
    task_config = config.copy()
    if 'cumulus_message' in task_config:
        del task_config['cumulus_message']
    write_error('Ending resolve_config_templates')
    return _resolve_config_object(event, task_config)


def store_remote_response(incoming_event, default_max_size, config_keys):
    """
    * Stores part of a response message in S3 if it is too big to send to StepFunctions
    * @param {*} incoming_event    - The response message
    * @param {*} default_max_size  - The maximum size (in bytes) a response message portion
    *                                can be before the method will store it in s3
    * @param {*} config_keys       - A list of valid CMA configuration keys
    * @returns {*} A response message, possibly referencing an S3 object for its contents
    """
    write_error('Starting store_remote_response')
    event = deepcopy(incoming_event)
    replace_config = event.get('ReplaceConfig', None)
    if not replace_config:
        return event
    # Set default value if FullMessage flag set
    if replace_config.get('FullMessage', False):
        replace_config['Path'] = '$'

    replace_config_values = _parse_remote_config_from_event(replace_config, default_max_size)

    for key in config_keys:
        if event.get(key):
            del event[key]

    cumulus_meta = deepcopy(event['cumulus_meta'])
    replacement_data = replace_config_values['parsed_json_path'].find(event)
    if len(replacement_data) != 1:
        raise Exception(f'JSON path invalid: {replace_config_values["parsed_json_path"]}')
    replacement_data = replacement_data[0]

    estimated_data_size = len(json.dumps(replacement_data.value).encode(('utf-8')))

    if estimated_data_size < replace_config_values['max_size']:
        return event

    _s3 = s3()
    s3_bucket = event['cumulus_meta']['system_bucket']
    s3_key = ('/').join(['events', str(uuid.uuid4())])
    s3_params = {
        'Expires': datetime.utcnow() + timedelta(days=7),  # Expire in a week
        'Body': json.dumps(replacement_data.value)
    }
    _s3.Object(s3_bucket, s3_key).put(**s3_params)

    try:
        replacement_data.value.clear()
    except AttributeError:
        replace_config_values['parsed_json_path'].update(event, '')

    remote_configuration = {'Bucket': s3_bucket, 'Key': s3_key,
                            'TargetPath': replace_config_values['target_path']}
    event['cumulus_meta'] = event.get('cumulus_meta', cumulus_meta)
    event['replace'] = remote_configuration
    write_error('store_remote_response')
    return event

def _resolve_config_object(event, config):
    """
    * Recursive helper for resolve_config_templates
    *
    * Given a config object containing possible JSONPath-templated values, resolves
    * all the values in the object using JSONPaths into the provided event.
    *
    * @param {*} event The event that paths resolve against
    * @param {*} config A config object, containing paths
    * @returns {*} A config object with all JSONPaths resolved
    """

    if isinstance(config, str):
        return resolve_path_str(event, config)

    if isinstance(config, list):
        for i in range(0, len(config)):  # pylint: disable=consider-using-enumerate
            config[i] = _resolve_config_object(event, config[i])
        return config

    if (config is not None and isinstance(config, dict)):
        result = {}
        for key in config.keys():
            result[key] = _resolve_config_object(event, config[key])
        return result

    return config


def _parse_remote_config_from_event(replace_config, default_max_size):
    source_path = replace_config['Path']
    target_path = replace_config.get('TargetPath', replace_config['Path'])
    default_max_size = replace_config.get('MaxSize', default_max_size)
    parsed_json_path = parse(source_path)

    return {
        'target_path': target_path,
        'max_size': default_max_size,
        'parsed_json_path': parsed_json_path,
    }



import os
import json

from copy import deepcopy
from jsonschema import validate

from .util import assign_json_path_value
from .cumulus_message import (resolve_config_templates, resolve_input,
                              resolve_path_str, load_config, load_remote_event,
                              store_remote_response)


class MessageAdapter:
    """
    transforms the cumulus message
    """
    REMOTE_DEFAULT_MAX_SIZE = 0
    CMA_CONFIG_KEYS = ['ReplaceConfig', 'task_config']

    def __init__(self, schemas=None):
        self.schemas = schemas

    ##################################
    #  Input message interpretation  #
    ##################################

    @staticmethod
    def __parse_parameter_configuration(event):
        parsed_event = event
        if event.get('cma'):
            updated_event = {k: v for (k, v) in event['cma'].items() if k != 'event'}
            parsed_event = event['cma']['event']
            parsed_event.update(updated_event)
        return parsed_event

    def load_and_update_remote_event(self, incoming_event, context):
        """
        * Looks at a Cumulus message. If the message has part of its data stored remotely in
        * S3, fetches that data, otherwise it returns the full message, both cases updated with
        * task metadata.
        * If event uses parameterized configuration, converts message into a
        * Cumulus message and ensures that incoming parameter keys are not overridden
        * @param {*} event The input Lambda event in the Cumulus message protocol
        * @returns {*} the full event data
        """
        event = deepcopy(incoming_event)

        if incoming_event.get('cma'):
            cma_event = deepcopy(incoming_event)
            event = load_remote_event(event['cma'].get('event'))
            cma_event['cma']['event'].update(event)
            event = self.__parse_parameter_configuration(cma_event)
        else:
            event = load_remote_event(event)

        if context and 'meta' in event:
            task_meta = {}
            task_meta['name'] = context.get('function_name', context.get('functionName'))
            task_meta['version'] = context.get('function_version', context.get('functionVersion'))
            task_meta['arn'] = context.get('invoked_function_arn',
                                           context.get('invokedFunctionArn',
                                                       context.get('activityArn')))
            if not 'workflow_tasks' in event['meta']:
                event['meta']['workflow_tasks'] = {}
            task_index = len(event['meta']['workflow_tasks'])
            event['meta']['workflow_tasks'][task_index] = task_meta
        return event

    def __get_jsonschema(self, schema_type):
        schemas = self.schemas
        root_dir = os.environ.get("LAMBDA_TASK_ROOT", '')
        has_schema = schemas and schemas.get(schema_type)
        rel_filepath = schemas.get(schema_type) if has_schema else f'schemas/{schema_type}.json'
        filepath = os.path.join(root_dir, rel_filepath)
        return filepath if os.path.exists(filepath) else None

    def __validate_json(self, document, schema_type):
        """
        check that json is valid based on a schema
        """
        schema_filepath = self.__get_jsonschema(schema_type)
        if schema_filepath:
            with open(schema_filepath, encoding='utf-8') as schema_handle:
                schema = json.load(schema_handle)
                try:
                    validate(document, schema)
                except Exception as exception:
                    exception.message = f'{schema_type} schema: {str(exception)}'
                    raise exception

    def load_nested_event(self, event):
        """
        * Interprets an incoming event as a Cumulus workflow message
        *
        * @param {*} event The input message sent to the Lambda
        * @returns {*} message that is ready to pass to an inner task
        """
        config = load_config(event)
        final_config = resolve_config_templates(event, config)
        final_payload = resolve_input(event, config)
        response = {'input': final_payload}
        self.__validate_json(final_payload, 'input')
        if final_config:
            self.__validate_json(final_config, 'config')
            response['config'] = final_config
        else:
            response['config'] = {}
        if 'cumulus_message' in config:
            response['messageConfig'] = config['cumulus_message']

        # add cumulus_config property, only selective attributes from event.cumulus_meta are added
        if 'cumulus_meta' in event:
            response['cumulus_config'] = {}
            # add both attributes or none of them
            attributes = ['state_machine', 'execution_name']
            if all(attribute in event['cumulus_meta'] for attribute in attributes):
                for attribute in attributes:
                    response['cumulus_config'][attribute] = event['cumulus_meta'][attribute]

            # add attribute cumulus_context
            if 'cumulus_context' in event['cumulus_meta']:
                cumulus_context = event['cumulus_meta']['cumulus_context']
                response['cumulus_config']['cumulus_context'] = cumulus_context

            if not response['cumulus_config']:
                del response['cumulus_config']

        return response

    @staticmethod
    def __assign_outputs(handler_response, event, message_config):
        """
        * Applies a task's return value to an output message as defined in config.cumulus_message
        *
        * @param {*} handler_response The task's return value
        * @param {*} event The output message to apply the return value to
        * @param {*} messageConfig The cumulus_message configuration
        * @returns {*} The output message with the nested response applied
        """
        result = deepcopy(event)
        if message_config is not None and 'outputs' in message_config:
            outputs = message_config['outputs']
            result['payload'] = {}
            for output in outputs:
                source_path = output['source']
                dest_path = output['destination']
                dest_json_path = dest_path.lstrip('{').rstrip('}')
                value = resolve_path_str(handler_response, source_path)
                result = assign_json_path_value(result, dest_json_path, value)
        else:
            result['payload'] = handler_response

        return result

    def create_next_event(self, handler_response, event, message_config):
        """
        * Creates the output message returned by a task
        *
        * @param {*} handler_response The response returned by the inner task code
        * @param {*} event The input message sent to the Lambda
        * @param {*} message_config The cumulus_message object configured for the task
        * @returns {*} the output message to be returned
        """
        self.__validate_json(handler_response, 'output')

        result = self.__assign_outputs(handler_response, event, message_config)
        if not result.get('exception'):
            result['exception'] = 'None'
        if 'replace' in result:
            del result['replace']
        return store_remote_response(result, self.REMOTE_DEFAULT_MAX_SIZE, self.CMA_CONFIG_KEYS)



from copy import deepcopy
from jsonpath_ng import parse


def assign_json_path_value(source_message, jspath, value):
    """
    * Assign (update or insert) a value to message based on jsonpath.
    * Create the keys if jspath doesn't already exist in the message. In this case, we
    * support 'simple' jsonpath like $.path1.path2.path3....
    * @param {dict} source_message The message to be updated
    * @param {string} jspath JSON path string
    * @param {*} value Value to update to
    * @return {*} updated message
    """
    message = deepcopy(source_message)
    if not parse(jspath).find(message):
        paths = jspath.lstrip('$.').split('.')
        current_item = message
        key_not_found = False
        for path in paths:
            if key_not_found or path not in current_item:
                key_not_found = True
                new_path_dict = {}
                # Add missing key to existing dict
                current_item[path] = new_path_dict
                # Set current item to newly created dict
                current_item = new_path_dict
            else:
                current_item = current_item[path]
    parse(jspath).update(message, value)
    return message



""" Determines the correct AWS endpoint for AWS services """
import os
from boto3 import resource

def localhost_s3_url():
    """ Returns configured LOCALSTACK_HOST url or default for localstack s3 """
    if 'LOCALSTACK_HOST' in os.environ:
        s3_url = f"http://{os.environ['LOCALSTACK_HOST']}:4572"
    else:
        s3_url = 'http://localhost:4572'
    return s3_url


def s3():
    """ Determines the endpoint for the S3 service """

    if ('CUMULUS_ENV' in os.environ) and (os.environ['CUMULUS_ENV'] == 'testing'):
        return resource(
            service_name='s3',
            endpoint_url=localhost_s3_url(),
            aws_access_key_id='my-id',
            aws_secret_access_key='my-secret',
            region_name='us-east-1',
            verify=False
        )
    return resource('s3')

def _get_sfn_execution_arn_by_name(state_machine_arn, execution_name):
    """
    * Given a state machine arn and execution name, returns the execution's ARN
    * @param {string} state_machine_arn The ARN of the state machine containing the execution
    * @param {string} execution_name The name of the execution
    * @returns {string} The execution's ARN
    """
    return (':').join([state_machine_arn.replace(':stateMachine:', ':execution:'),
                       execution_name])


def _get_task_name_from_execution_history(execution_history, arn):
    """
    * Given an execution history object returned by the StepFunctions API and an optional
    * Activity or Lambda ARN returns the most recent task name started for the given ARN,
    * or if no ARN is supplied, the most recent task started.
    *
    * IMPORTANT! If no ARN is supplied, this message assumes that the most recently started
    * execution is the desired execution. This WILL BREAK parallel executions, so always supply
    * this if possible.
    *
    * @param {dict} executionHistory The execution history returned by getExecutionHistory,
    * assumed to be sorted so most recent executions come last
    * @param {string} arn An ARN to an Activity or Lambda to find. See "IMPORTANT!"
    * @throws If no matching task is found
    * @returns {string} The matching task name
    """
    events_by_id = {}

    # Create a lookup table for finding events by their id
    for event in execution_history['events']:
        events_by_id[event['id']] = event

    for step in execution_history['events']:
        # Find the ARN in the history (the API is awful here).  When found, return its
        # previousEventId's (TaskStateEntered) name
        lambda_of_type_and_matching_arn = (
            (step['type'] == 'LambdaFunctionScheduled' and
             step['lambdaFunctionScheduledEventDetails']['resource'] == arn) or
            (step['type'] == 'ActivityScheduled' and
             step['activityScheduledEventDetails']['resource'] == arn))

        if (arn is not None and lambda_of_type_and_matching_arn and
                'stateEnteredEventDetails' in events_by_id[step['previousEventId']]):
            return events_by_id[step['previousEventId']]['stateEnteredEventDetails']['name']

        if step['type'] == 'TaskStateEntered':
            return step['stateEnteredEventDetails']['name']

    raise LookupError('No task found for ' + arn)



"""A setuptools based setup module.
See:
https://packaging.python.org/en/latest/distributing.html
https://github.com/pypa/sampleproject
"""
# pylint: disable=W0402
import imp

from os import path
# Always prefer setuptools over distutils
from setuptools import setup, find_packages

here = path.abspath(path.dirname(__file__))

# get the dependencies and installs
with open(path.join(here, 'requirements.txt'), encoding='utf-8') as f:
    all_reqs = f.read().split('\n')
install_requires = [x.strip() for x in all_reqs if 'git+' not in x]
dependency_links = [x.strip().replace('git+', '') for x in all_reqs if 'git+' in x]
print(install_requires)
print(dependency_links)

# Get the long description from the README file
with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

# Arguments marked as "Required" below must be included for upload to PyPI.
# Fields marked as "Optional" may be commented out.
__version__ = imp.load_source('version', 'message_adapter/version.py').__version__

setup(
    name='cumulus-message-adapter',  # Required

    # Versions should comply with PEP 440:
    # https://www.python.org/dev/peps/pep-0440/
    version=__version__,  # Required

    # This is a one-line description or tagline of what your project does. This
    # corresponds to the "Summary" metadata field:
    # https://packaging.python.org/specifications/core-metadata/#summary
    description=('A command-line interface for preparing and outputting '
                 'Cumulus Messages for Cumulus Tasks'),  # Required
    long_description=long_description,  # Optional
    url='https://github.com/nasa/cumulus-message-adapter',  # Optional
    author='Cumulus Authors',  # Optional
    author_email='info@developmentseed.org',  # Optional
    classifiers=[  # Optional
        # Indicate who your project is intended for
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8'
    ],
    keywords='nasa cumulus message adapter',  # Optional
    packages=find_packages(exclude=['.circleci', 'contrib', 'docs', 'tests']),  # Required
    install_requires=install_requires,
    python_requires='~=3.6',
    dependency_links=dependency_links
)
